/*

-Procedure turnfast ( Determine the kinematic parameters
                      of the fastest possible turn within
                      given angular velocity and acceleration
                      constraints. )

   Jeff Bytof, last update 26-September-2001

*/

#include <math.h>

#include "SpiceUsr.h"
#include "SpiceZmc.h"
#include "SpiceZfc.h"

void turnfast ( SpiceDouble     convel     [3],
                SpiceDouble     conacc     [3],
              
                SpiceDouble     turn_axis  [3],
                SpiceDouble     turn_angle,
                SpiceDouble   * tout,
                SpiceDouble   * tbout,
                SpiceDouble   * tcout,
                SpiceDouble     alpha      [3],
                SpiceDouble     omega      [3],
 
                SpiceBoolean  * found,
                SpiceInt      * ier            )

{

/*

-Brief_I/O

    Variable    I/O      Description
    --------    ---      -----------
    convel       I       Angular velocity constraint vector (rad/sec). 
    convel       I       Angular acceleration constraint vector (rad/sec^2).

    turn_axis    I       Turn axis vector. 
    turn_angle   I       Turn angle. 

    tout         O       Total duration of turn, in seconds.
    tbout        O       Duration of burn, in seconds.
    tcout        O       Duration of coast, in seconds.
    alpha        O       Angular acceleration vector (rad/sec^2).
    omega        O       Angular velocity vector (rad/sec).

    found        O       Logical flag, SPICETRUE if no error is found.
    ier          O       If found = SPICEFALSE returns an error code.

*/



/*
    Local variables.
*/


   SpiceDouble         afac;
   SpiceDouble         cfact; 
   SpiceDouble         maxacc;
   SpiceDouble         maxvel;
   SpiceDouble         t; 
   SpiceDouble         tb; 
   SpiceDouble         tc; 
   SpiceDouble         theta;
   SpiceDouble         TOL;
   SpiceDouble         vfac;
   SpiceDouble         vsep;

   SpiceInt            i;
   SpiceInt            ier2; 

   SpiceBoolean        fnd;



/*
   Set values.
*/

   TOL = 1.e-8;
   cfact = 0.001;

   *found = SPICEFALSE;
   *ier   = 0;
   *tout  = 0.0;
   *tbout = 0.0;
   *tcout = 0.0;

   theta = turn_angle;


/*
   Pick out the maximum angular velcity and 
   acceleration components and use them in 
   the initial estimate of the turn parameters.
*/

   maxvel = maxd_c ( (SpiceInt) 3, convel[0], convel[1], convel[2] );
   maxacc = maxd_c ( (SpiceInt) 3, conacc[0], conacc[1], conacc[2] );

/*
   If turn angle is zero, return all zeroes.
*/

   if ( fabs(theta) == 0.0 ) { 
      vpack_c ( 0.0, 0.0, 0.0, alpha );
      vpack_c ( 0.0, 0.0, 0.0, omega );
      *tout  = 0.0;
      *tbout = 0.0;
      *tcout = 0.0;
      *found = SPICETRUE;
      return;
   }


/*
   Estimate the total turn duration, accelerating
   the first half and decelerating the last half. 
   A short coast period is included to avoid numerical
   problems.
*/

   tb = sqrt ( fabs(theta) / ( maxacc * ( 1.0 + cfact ) ) );
   tc = tb * cfact;
   t  = (2.0 * tb) + tc;

/*
   Compute turn parameters.
*/

   turnpar ( turn_axis,
             turn_angle, 
             t, 
             maxacc,
            &tb, 
            &tc, 
             alpha,
             omega, 
            &fnd,     
            &ier2 ); 

   if ( fnd != SPICETRUE ) { 
      *ier = 1000 + ier2;
      return;
   }

/*
   Scale estimated angular acceleration to fit
   constraint vector.  
*/

   afac = 0.0;
 
   for ( i=0; i<3; i++ ) {
      afac = maxd_c ( (SpiceInt) 2, afac, fabs( alpha[i]/conacc[i] ));
   }


   maxacc = maxacc / afac;

   for ( i=0; i<3; i++ ) {
      alpha[i] = alpha[i] / afac;
   }

/*
   Revise timing parameters given the changed acceleration.
*/

   tb = sqrt( fabs(theta) / (maxacc * ( 1.0 + cfact )));
   tc = tb * cfact;
   t  = ( 2.0 * tb ) + tc;

/*
   Revise turn parameters.
*/

   turnpar ( turn_axis, 
             turn_angle, 
             t, 
             maxacc,
            &tb, 
            &tc, 
             alpha,
             omega, 
            &fnd,     
            &ier2 ); 


   if ( fnd != SPICETRUE ) { 
      *ier = 2000 + ier2;
      return;
   }

/*
   Scale estimated angular velocity to fit constraint vector.
*/

   vfac = 1.0;
  
   for ( i=0; i<3; i++ ) {
      vfac = maxd_c ( (SpiceInt) 2, vfac, fabs( omega[i] / convel[i] ));
   } 

/*
   Return now if no constraints are violated.

   if ( vfac == 1.0 ) {
      *tbout = tb;
      *tcout = tc;
      *found = SPICETRUE;
      return; 
   }
      

/* 
   Revise timing parameters given the changed angular velocity.
*/


   tb = tb / vfac;
   tc = tb * (( ( 1.0 + cfact ) * ( vfac*vfac )) - 1.0 );
   t  = ( 2.0 * tb ) + tc;

   for ( i=0; i<3; i++ ) {
      omega[i] = alpha[i] * tb;
   }


   *tout  = t;
   *tbout = tb;
   *tcout = tc;

   *found = SPICETRUE;

   return; 

}
